function G(id) {
	return document.getElementById(id);
}

var Remote = (function() {

	var createXMLHttpRequestPromise = function(options) {
		return new Promise(function(resolve, reject) {
			var xhr = new XMLHttpRequest();
			xhr.open(options.method || 'GET', options.url, options.async || true);

			// Set request headers if options.headers provided.
			Object.keys(options.headers || {}).forEach(function (key) {
				xhr.setRequestHeader(key, options.headers[key]);
			});

			xhr.onreadystatechange = function() {
				if(xhr.readyState !== 4) {
					return;
				}

				if([200, 304].indexOf(xhr.status) === -1) {
				  reject(options.url + ' responded with a status of ' + xhr.status);
				} else {
				  resolve(xhr.responseText);
				}
			};

			xhr.send(JSON.stringify(options.data) || void 0);
		});
	}

	var request = function(node, onSuccess, onError) {
		var options = {
			url : "/" + node,
			method : 'GET'
		}

		createXMLHttpRequestPromise(options)
			.then(onSuccess || function() {})
			.catch(onError || function() {});
	}

	var get = function(url, async) {
		var options = {
			url : url,
			async : async,
			method : 'GET'
		};

		return createXMLHttpRequestPromise(options);
	}

	var post = function(url, data, async) {
		var options = {
			url : url,
			data : data,
			async : async,
			method : 'POST',
			headers : {
				'Content-Type' : 'application/json;charset=UTF-8'
			}
		};

		return createXMLHttpRequestPromise(options);
	}

	

	return {
		request : request,
		get : get,
		post : post
	}

})();

var Console = (function() {

	var consoleBuffer = {
		lines : 0,
		maxLines: 200
	};

	var commandHistory = {
		commands: [],
		index : -1,
		maxCommands : 100
	};

	function printLine(text) {
		var buffer = document.getElementById("commandbuffer");
		var scrollFollow = buffer.scrollHeight - buffer.scrollTop === buffer.clientHeight;
		var node = document.createElement("div");
		var prenode = document.createElement("pre");
		prenode.innerHTML = text.replace(/</g, '&lt;').replace(/>/g, '&gt;').replace(/\n/g, '<br />').replace(/  /g, '&nbsp;&nbsp;');
		node.appendChild(prenode);
		buffer.appendChild(node);
		consoleBuffer.lines++;
		if (consoleBuffer.lines >= consoleBuffer.maxLines) {
			buffer.removeChild(buffer.childNodes[0]);
		}
		if (scrollFollow) {
			buffer.scrollTop = buffer.scrollHeight - buffer.clientHeight;
		}
	}

	function getCommandElement() {
		return document.getElementById("command");
	}

	function getCommandText() {
		return getCommandElement().value;
	}

	function setCommandText(text) {
		var cmd = getCommandElement();
		cmd.value = text;
		cmd.setSelectionRange(text.length, text.length);
	}

	function commonPrefix(commands){
		commands = commands.sort();
		var first = commands[0], last = commands[commands.length - 1];
		var i = 0;
		while (i < first.length && first.charAt(i) === last.charAt(i)) ++i;
		return first.substring(0, i);
	}

	function updateCommandFromHistory() {
		if (commandHistory.index >= commandHistory.commands.length) {
			commandHistory.index = commandHistory.commands.length-1;
		}
		if (commandHistory.index < -1) {
			commandHistory.index = -1;
		}

		if (commandHistory.index > -1) {
			setCommandText(commandHistory.commands[commandHistory.index]);
		} else {
			setCommandText("");
		}
	}

	function request(command, onSuccess) {
		Remote.get("console/command/" + command)
			.then(onSuccess)
			.catch(printLine);
	}

	function parseCommands(response) {
		var lines = JSON.parse(response);
		var commands = [];
		for(key in lines) {
			if (key != "command" && key != "line 0") {
				commands.push(lines[key].trim());
			}
		}
		return commands;
	}

	function ifCommandExistsThenElse(command, doThen, doElse) {
		request("help", function(response) {
			var commands = parseCommands(response);
			if (commands.indexOf(command) != -1)
				doThen();
			else
				doElse();
		});
	}

	function ifCommandExistsThen(command, doThen) {
		ifCommandExistsThenElse(command, doThen, function() { });
	}

	function updateConsoleVariableIndicator(indicator, variable, suffix) {
		Console.request(variable, function(response) {
			parsedResponse = JSON.parse(response);
			value = parsedResponse["line 0"].split(": ")[1]
			document.getElementById(indicator).innerHTML = value + suffix;
		})
	}

	function printWelcome() {
		printLine("Welcome to the Fiction Factory Web Console")
		printLine("* to list the available commands enter: help")
		printLine("* to autocomplete a command press TAB")
		printLine("* to see all autocompletion options press TAB twice")
	}

	function init() {
		printWelcome();
	}

	function onKeyDown(event) {
		if (event.keyCode == 38) {
			event.preventDefault();
			if (commandHistory.commands.length > 0) {
				commandHistory.index++;
				updateCommandFromHistory();
			}
		} else if(event.keyCode == 40) {
			event.preventDefault();
			if (commandHistory.commands.length > 0) {
				commandHistory.index--;
				updateCommandFromHistory();
			}
		} else if (event.keyCode == 13) {
			var command = getCommandText().trim();
			if (command == "")
				return;
			commandHistory.index = -1;
			commandHistory.commands.unshift(command);
			if (commandHistory.commands.length > commandHistory.maxCommands) {
				commandHistory.pop();
			}
			setCommandText("");
			printLine("$ " + command);
			request(command, function(response) {
				var lines = JSON.parse(response);
				for(key in lines) {
					if (key != "command") {
						printLine(lines[key]);
					}
				}
			});
		} else if (event.keyCode == 9) {
			event.preventDefault();
			var commandElements = getCommandText().split(" ");
			if (commandElements.length > 1)
				return;
			var command = commandElements[0];
			if (!command || command.trim() == "")
				return;
			command = command.trim();
			request("help", function(response) {
				var commands = parseCommands(response);

				commands = commands.filter(function(c) { return c.startsWith(command); });
				if (commands.length == 0)
					return;
				command = commonPrefix(commands);
				if (getCommandText() == command && commands.length > 1) {
					printLine(commands.join(" "));
					return;
				}
				setCommandText(command);
			});
		}
	}

	function enableImageBackground() {
		document.body.className = "image-background";
	}

	return {
		init : init,
		onKeyDown: onKeyDown,
		request : request,
		ifCommandExistsThen: ifCommandExistsThen,
		ifCommandExistsThenElse: ifCommandExistsThenElse,
		enableImageBackground: enableImageBackground
	};

})();

var CustomContent = (function() {
	function load(file, dest, callback) {
		Remote.request(file, function(text) {
			if (typeof dest === 'string')
				G(dest).innerHTML = text;
			else if (typeof dest === 'function')
				dest.call(null, text);
			if (typeof callback === 'function')
				callback.call(null, text);

		});
	}
	return {
		load : load
	}

})();

var NodeBooleanIndicator = function(elementName, node) {
	return {
		refresh: function() {
			Remote.request(node, function(response) {
				document.getElementById(elementName).className = (response == "true") ? "indicator_on" : "indicator_off";
			})
		}
	}
}

var CVarIndicator = function(elementName, variable, suffix) {
	return {
		refresh: function() {
			Console.request(variable, function(response) {
				parsedResponse = JSON.parse(response);
				value = parsedResponse["line 0"].split(": ")[1]
				document.getElementById(elementName).innerHTML = value + suffix;
			})
		}
	}
}


window.addEventListener('load', function() {

	Console.init();

	document.getElementById("command").focus();
	document.getElementById("command").addEventListener("keydown", Console.onKeyDown);
	document.getElementById("commandbuffer").addEventListener("click", function() {document.getElementById("command").focus();});
});
